@@ -8,6 +8,7 @@ from collections import OrderedDict
 import numpy as np
 
 from . import display
+from . import gradients as grad
 from . import losses as losses_module
 from . import metrics as metrics_module
 from . import optimizers
@@ -17,11 +18,11 @@ from .callbacks import CallbackList
 
 
 class Model(object):
-    """A ``Model`` trains a ``Map`` on a ``Data``.
+    """A ``Model`` trains a ``NN`` on a ``Data``.
 
     Args:
         data: ``deepxde.data.Data`` instance.
-        net: ``deepxde.maps.Map`` instance.
+        net: ``deepxde.nn.NN`` instance.
     """
 
     def __init__(self, data, net):
@@ -39,7 +40,7 @@ class Model(object):
 
         # Backend-dependent attributes
         # Tensor or callable
-        self.losses = None
+        self.outputs = None
         self.outputs_losses = None
         self.train_step = None
         if backend_name == "tensorflow.compat.v1":
@@ -85,14 +86,8 @@ class Model(object):
         """
         print("Compiling model...")
 
-        if backend_name == "tensorflow.compat.v1":
-            if not self.net.built:
-                self.net.build()
-            if self.sess is None:
-                self.sess = tf.Session()
-                self.saver = tf.train.Saver(max_to_keep=None)
-
         self.opt_name = optimizer
+        loss_fn = losses_module.get(loss)
         if external_trainable_variables is None:
             self.external_trainable_variables = []
         else:
@@ -106,107 +101,188 @@ class Model(object):
                 external_trainable_variables = [external_trainable_variables]
             self.external_trainable_variables = external_trainable_variables
 
-        loss_fn = losses_module.get(loss)
         if backend_name == "tensorflow.compat.v1":
+            self._compile_tensorflow_compat_v1(lr, loss_fn, decay, loss_weights)
+        elif backend_name == "tensorflow":
+            self._compile_tensorflow(lr, loss_fn, decay, loss_weights)
+        elif backend_name == "pytorch":
+            self._compile_pytorch(lr, loss_fn, decay, loss_weights)
+
+        # metrics may use model variables such as self.net, and thus are instantiated
+        # after backend compile.
+        metrics = metrics or []
+        self.metrics = [metrics_module.get(m) for m in metrics]
+
+    def _compile_tensorflow_compat_v1(self, lr, loss_fn, decay, loss_weights):
+        """tensorflow.compat.v1"""
+        if not self.net.built:
+            self.net.build()
+        if self.sess is None:
+            self.sess = tf.Session()
+            self.saver = tf.train.Saver(max_to_keep=None)
+
+        # Data losses
+        losses = self.data.losses(self.net.targets, self.net.outputs, loss_fn, self)
+        if not isinstance(losses, list):
+            losses = [losses]
+        # Regularization loss
+        if self.net.regularizer is not None:
+            losses.append(tf.losses.get_regularization_loss())
+        losses = tf.convert_to_tensor(losses)
+        # Weighted losses
+        if loss_weights is not None:
+            losses *= loss_weights
+            self.losshistory.set_loss_weights(loss_weights)
+        total_loss = tf.math.reduce_sum(losses)
+
+        # Tensors
+        self.outputs = self.net.outputs
+        self.outputs_losses = [self.net.outputs, losses]
+        self.train_step = optimizers.get(
+            total_loss, self.opt_name, learning_rate=lr, decay=decay
+        )
+
+    def _compile_tensorflow(self, lr, loss_fn, decay, loss_weights):
+        """tensorflow"""
+
+        # TODO: Avoid creating multiple graphs by using tf.TensorSpec.
+        @tf.function
+        def outputs(training, inputs):
+            return self.net(inputs, training=training)
+
+        # TODO: Avoid creating multiple graphs by using tf.TensorSpec.
+        @tf.function
+        def outputs_losses(training, inputs, targets, auxiliary_vars):
+            # TODO: Add training
+            # self.net.training = training
+            self.net.inputs = inputs
+            self.net.targets = targets
+            self.net.auxiliary_vars = auxiliary_vars
+            # Don't call outputs() decorated by @tf.function above, otherwise the
+            # gradient of outputs wrt inputs will be lost here.
+            outputs_ = self.net(inputs, training=training)
             # Data losses
-            losses = self.data.losses(self.net.targets, self.net.outputs, loss_fn, self)
+            losses = self.data.losses(targets, outputs_, loss_fn, self)
             if not isinstance(losses, list):
                 losses = [losses]
             # Regularization loss
             if self.net.regularizer is not None:
-                losses.append(tf.losses.get_regularization_loss())
+                losses += [tf.math.reduce_sum(self.net.losses)]
             losses = tf.convert_to_tensor(losses)
-            # Weighted losses
+            # TODO: Weighted losses
             if loss_weights is not None:
-                losses *= loss_weights
-                self.losshistory.set_loss_weights(loss_weights)
-            total_loss = tf.math.reduce_sum(losses)
-            # Tensors
-            self.losses = losses
-            self.outputs_losses = [self.net.outputs, losses]
-            self.train_step = optimizers.get(
-                total_loss, self.opt_name, learning_rate=lr, decay=decay
-            )
-        elif backend_name == "tensorflow":
-
-            def compute_losses(targets, outputs):
-                # Data losses
-                losses = self.data.losses(targets, outputs, loss_fn, self)
-                if not isinstance(losses, list):
-                    losses = [losses]
-                if self.net.regularizer is not None:
-                    losses += [tf.math.reduce_sum(self.net.losses)]
-                losses = tf.convert_to_tensor(losses)
-                # TODO: Weighted losses
-                return losses
-
-            # TODO: Avoid creating multiple graphs by using tf.TensorSpec.
-            @tf.function
-            def outputs_losses(data_id, inputs, targets, auxiliary_vars=None):
-                self.net.data_id = data_id
-                self.net.inputs = inputs
-                self.net.targets = targets
-                self.net.auxiliary_vars = auxiliary_vars
-                outputs = self.net(inputs)
-                losses = compute_losses(targets, outputs)
-                return outputs, losses
-
-            opt = optimizers.get(self.opt_name, learning_rate=lr, decay=decay)
-
-            @tf.function
-            def train_step(data_id, inputs, targets, auxiliary_vars=None):
-                with tf.GradientTape() as tape:
-                    _, losses = outputs_losses(data_id, inputs, targets, auxiliary_vars)
-                    total_loss = tf.math.reduce_sum(losses)
-                trainable_variables = (
-                    self.net.trainable_variables + self.external_trainable_variables
+                raise NotImplementedError(
+                    "Backend tensorflow doesn't support loss_weights"
                 )
-                grads = tape.gradient(total_loss, trainable_variables)
-                opt.apply_gradients(zip(grads, trainable_variables))
+            return outputs_, losses
 
-            # Callables
-            self.losses = compute_losses
-            self.outputs_losses = outputs_losses
-            self.train_step = train_step
+        opt = optimizers.get(self.opt_name, learning_rate=lr, decay=decay)
 
-        elif backend_name == "pytorch":
-
-            def compute_losses(targets, outputs):
-                # Data losses
-                losses = self.data.losses(targets, outputs, loss_fn, self)
-                if not isinstance(losses, list):
-                    losses = [losses]
-                # TODO: regularization
-                # TODO: Weighted losses
-                losses = torch.stack(losses)
-                return losses
-
-            def outputs_losses(data_id, inputs, targets):
-                # Is this the best place to convert?
-                inputs = torch.from_numpy(inputs)
-                targets = torch.from_numpy(targets)
-                outputs = self.net(inputs)
-                losses = compute_losses(targets, outputs)
-                return outputs, losses
-
-            opt = optimizers.get(
-                self.net.parameters(), self.opt_name, learning_rate=lr, decay=decay
+        @tf.function
+        def train_step(inputs, targets, auxiliary_vars):
+            # inputs and targets are np.ndarray and automatically converted to Tensor.
+            with tf.GradientTape() as tape:
+                losses = outputs_losses(True, inputs, targets, auxiliary_vars)[1]
+                total_loss = tf.math.reduce_sum(losses)
+            trainable_variables = (
+                self.net.trainable_variables + self.external_trainable_variables
             )
+            grads = tape.gradient(total_loss, trainable_variables)
+            opt.apply_gradients(zip(grads, trainable_variables))
 
-            def train_step(data_id, inputs, targets):
-                _, losses = outputs_losses(data_id, inputs, targets)
-                total_loss = torch.sum(losses)
-                opt.zero_grad()
-                total_loss.backward()
-                opt.step()
+        # Callables
+        self.outputs = outputs
+        self.outputs_losses = outputs_losses
+        self.train_step = train_step
 
-            # Callables
-            self.losses = compute_losses
-            self.outputs_losses = outputs_losses
-            self.train_step = train_step
+    def _compile_pytorch(self, lr, loss_fn, decay, loss_weights):
+        """pytorch"""
 
-        metrics = metrics or []
-        self.metrics = [metrics_module.get(m) for m in metrics]
+        def outputs(inputs):
+            # TODO: Add training
+            # TODO: Use torch.no_grad() if training is False
+            inputs = torch.from_numpy(inputs)
+            inputs.requires_grad_()
+            self.net.inputs = inputs
+            return self.net(inputs)
+
+        def outputs_losses(inputs, targets):
+            outputs_ = outputs(inputs)
+            # Data losses
+            if targets is not None:
+                targets = torch.from_numpy(targets)
+            losses = self.data.losses(targets, outputs_, loss_fn, self)
+            if not isinstance(losses, list):
+                losses = [losses]
+            # TODO: regularization
+            # TODO: Weighted losses
+            if loss_weights is not None:
+                raise NotImplementedError(
+                    "Backend pytorch doesn't support loss_weights"
+                )
+            losses = torch.stack(losses)
+            # Clear cached Jacobians and Hessians.
+            grad.clear()
+            return outputs_, losses
+
+        # Another way is using Per-parameter options
+        # https://pytorch.org/docs/stable/optim.html#per-parameter-options
+        trainable_variables = (
+            list(self.net.parameters()) + self.external_trainable_variables
+        )
+        opt = optimizers.get(
+            trainable_variables, self.opt_name, learning_rate=lr, decay=decay
+        )
+
+        def train_step(inputs, targets):
+            losses = outputs_losses(inputs, targets)[1]
+            total_loss = torch.sum(losses)
+            opt.zero_grad()
+            total_loss.backward()
+            opt.step()
+
+        # Callables
+        self.outputs = outputs
+        self.outputs_losses = outputs_losses
+        self.train_step = train_step
+
+    def _outputs(self, training, inputs):
+        if backend_name == "tensorflow.compat.v1":
+            feed_dict = self.net.feed_dict(training, inputs)
+            return self.sess.run(self.outputs, feed_dict=feed_dict)
+        if backend_name == "tensorflow":
+            outs = self.outputs(training, inputs)
+        elif backend_name == "pytorch":
+            # TODO: training
+            with torch.no_grad():
+                outs = self.outputs(inputs)
+        if isinstance(outs, (list, tuple)):
+            return [out.numpy() for out in outs]
+        return outs.numpy()
+
+    def _train_step(self, inputs, targets, auxiliary_vars):
+        if backend_name == "tensorflow.compat.v1":
+            feed_dict = self.net.feed_dict(True, inputs, targets, auxiliary_vars)
+            self.sess.run(self.train_step, feed_dict=feed_dict)
+        elif backend_name == "tensorflow":
+            self.train_step(inputs, targets, auxiliary_vars)
+        elif backend_name == "pytorch":
+            # TODO: auxiliary_vars
+            self.train_step(inputs, targets)
+
+    def _run(self, fetches, training, inputs, targets, auxiliary_vars):
+        """Runs one "step" of computation of tensors or callables in `fetches`."""
+        if backend_name == "tensorflow.compat.v1":
+            feed_dict = self.net.feed_dict(training, inputs, targets, auxiliary_vars)
+            return self.sess.run(fetches, feed_dict=feed_dict)
+        if backend_name == "tensorflow":
+            outs = fetches(training, inputs, targets, auxiliary_vars)
+            return None if outs is None else [out.numpy() for out in outs]
+        if backend_name == "pytorch":
+            # TODO: Use torch.no_grad() in _test() and predict()
+            # TODO: training, auxiliary_vars
+            outs = fetches(inputs, targets)
+            return None if outs is None else [out.detach().numpy() for out in outs]
 
     @utils.timing
     def train(
@@ -214,7 +290,6 @@ class Model(object):
         epochs=None,
         batch_size=None,
         display_every=1000,
-        uncertainty=False,
         disregard_previous_best=False,
         callbacks=None,
         model_restore_path=None,
@@ -223,15 +298,14 @@ class Model(object):
         """Trains the model for a fixed number of epochs (iterations on a dataset).
 
         Args:
-            epochs: Integer. Number of epochs to train the model.
+            epochs: Integer. Number of iterations to train the model. Note: It is the
+                number of iterations, not the number of epochs.
             batch_size: Integer or ``None``. If you solve PDEs via ``dde.data.PDE`` or
                 ``dde.data.TimePDE``, do not use `batch_size`, and instead use
                 `dde.callbacks.PDEResidualResampler
                 <https://deepxde.readthedocs.io/en/latest/modules/deepxde.html#deepxde.callbacks.PDEResidualResampler>`_,
                 see an `example <https://github.com/lululxvi/deepxde/blob/master/examples/diffusion_1d_resample.py>`_.
             display_every: Integer. Print the loss and metrics every this steps.
-            uncertainty: Boolean. If ``True``, use Monte-Carlo Dropout to estimate
-                uncertainty.
             disregard_previous_best: If ``True``, disregard the previous saved best
                 model.
             callbacks: List of ``dde.callbacks.Callback`` instances. List of callbacks
@@ -261,14 +335,14 @@ class Model(object):
         self.stop_training = False
         self.train_state.set_data_train(*self.data.train_next_batch(self.batch_size))
         self.train_state.set_data_test(*self.data.test())
-        self._test(uncertainty)
+        self._test()
         self.callbacks.on_train_begin()
         if optimizers.is_external_optimizer(self.opt_name):
-            self._train_scipy(display_every, uncertainty)
+            self._train_scipy(display_every)
         else:
             if epochs is None:
                 raise ValueError("No epochs for {}.".format(self.opt_name))
-            self._train_sgd(epochs, display_every, uncertainty)
+            self._train_sgd(epochs, display_every)
         self.callbacks.on_train_end()
 
         print("")
@@ -277,7 +351,7 @@ class Model(object):
             self.save(model_save_path, verbose=1)
         return self.losshistory, self.train_state
 
-    def _train_sgd(self, epochs, display_every, uncertainty):
+    def _train_sgd(self, epochs, display_every):
         for i in range(epochs):
             self.callbacks.on_epoch_begin()
             self.callbacks.on_batch_begin()
@@ -285,11 +359,7 @@ class Model(object):
             self.train_state.set_data_train(
                 *self.data.train_next_batch(self.batch_size)
             )
-            self._run(
-                self.train_step,
-                True,
-                True,
-                np.uint8(0),
+            self._train_step(
                 self.train_state.X_train,
                 self.train_state.y_train,
                 self.train_state.train_aux_vars,
@@ -298,7 +368,7 @@ class Model(object):
             self.train_state.epoch += 1
             self.train_state.step += 1
             if self.train_state.step % display_every == 0 or i + 1 == epochs:
-                self._test(uncertainty)
+                self._test()
 
             self.callbacks.on_batch_end()
             self.callbacks.on_epoch_end()
@@ -306,8 +376,7 @@ class Model(object):
             if self.stop_training:
                 break
 
-    def _train_scipy(self, display_every, uncertainty):
-        # TODO: backend tensorflow
+    def _train_scipy(self, display_every):
         def loss_callback(loss_train):
             self.train_state.epoch += 1
             self.train_state.step += 1
@@ -323,8 +392,6 @@ class Model(object):
         self.train_state.set_data_train(*self.data.train_next_batch(self.batch_size))
         feed_dict = self.net.feed_dict(
             True,
-            True,
-            0,
             self.train_state.X_train,
             self.train_state.y_train,
             self.train_state.train_aux_vars,
@@ -332,17 +399,15 @@ class Model(object):
         self.train_step.minimize(
             self.sess,
             feed_dict=feed_dict,
-            fetches=[self.losses],
+            fetches=[self.outputs_losses[1]],
             loss_callback=loss_callback,
         )
-        self._test(uncertainty)
+        self._test()
 
-    def _test(self, uncertainty):
+    def _test(self):
         self.train_state.y_pred_train, self.train_state.loss_train = self._run(
             self.outputs_losses,
-            False,
-            False,
-            np.uint8(0),
+            True,
             self.train_state.X_train,
             self.train_state.y_train,
             self.train_state.train_aux_vars,
@@ -350,34 +415,10 @@ class Model(object):
         self.train_state.y_pred_test, self.train_state.loss_test = self._run(
             self.outputs_losses,
             False,
-            False,
-            np.uint8(1),
             self.train_state.X_test,
             self.train_state.y_test,
             self.train_state.test_aux_vars,
         )
-        # UQ via dropout
-        if uncertainty:
-            # TODO: support multi outputs
-            # TODO: backend tensorflow
-            losses, y_preds = [], []
-            feed_dict = self.net.feed_dict(
-                False,
-                True,
-                1,
-                self.train_state.X_test,
-                self.train_state.y_test,
-                self.train_state.test_aux_vars,
-            )
-            for _ in range(1000):
-                y_pred_test_one, loss_one = self.sess.run(
-                    self.outputs_losses, feed_dict=feed_dict
-                )
-                losses.append(loss_one)
-                y_preds.append(y_pred_test_one)
-            self.train_state.loss_test = np.mean(losses, axis=0)
-            self.train_state.y_pred_test = np.mean(y_preds, axis=0)
-            self.train_state.y_std_test = np.std(y_preds, axis=0)
 
         if isinstance(self.train_state.y_test, (list, tuple)):
             self.train_state.metrics_test = [
@@ -405,22 +446,19 @@ class Model(object):
         self.callbacks = CallbackList(callbacks=callbacks)
         self.callbacks.set_model(self)
         self.callbacks.on_predict_begin()
-        # TODO: use self._run for tensorflow
-        # TODO: predict operator with auxiliary_vars
-        if backend_name == "tensorflow.compat.v1":
-            if operator is None:
-                op = self.net.outputs
-            else:
+        if operator is None:
+            y = self._outputs(False, x)
+        else:
+            # TODO: predict operator with auxiliary_vars
+            if backend_name == "tensorflow.compat.v1":
                 if utils.get_num_args(operator) == 2:
                     op = operator(self.net.inputs, self.net.outputs)
                 elif utils.get_num_args(operator) == 3:
                     op = operator(self.net.inputs, self.net.outputs, x)
-            y = self._run(op, False, False, np.uint8(2), x, None, None)
-        elif backend_name == "tensorflow":
-            # TODO: avoid creating the same graph every time predict is called
-            if operator is None:
-                y = self.net(x).numpy()
-            else:
+                y = self._run(op, False, x, None, None)
+            elif backend_name == "tensorflow":
+                # TODO: avoid creating the same graph every time predict is called
+                # TODO: use self._run for tensorflow
                 if utils.get_num_args(operator) == 2:
 
                     @tf.function
@@ -436,6 +474,11 @@ class Model(object):
                         return operator(inputs, y, x)
 
                 y = op(x).numpy()
+            elif backend_name == "pytorch":
+                # TODO
+                raise NotImplementedError(
+                    "Model.predict hasn't been implemented for backend pytorch."
+                )
         self.callbacks.on_predict_end()
         return y
 
@@ -447,7 +490,11 @@ class Model(object):
 
     def state_dict(self):
         """Returns a dictionary containing all variables."""
-        # TODO: backend tensorflow
+        # TODO: backend tensorflow, pytorch
+        if backend_name != "tensorflow.compat.v1":
+            raise NotImplementedError(
+                "state_dict hasn't been implemented for this backend."
+            )
         destination = OrderedDict()
         variables_names = [v.name for v in tf.global_variables()]
         values = self.sess.run(variables_names)
@@ -464,7 +511,11 @@ class Model(object):
                 If `protocol` is "pickle", save using the Python pickle module. Only
                 "tf.train.Saver" protocol supports ``restore()``.
         """
-        # TODO: backend tensorflow
+        # TODO: backend tensorflow, pytorch
+        if backend_name != "tensorflow.compat.v1":
+            raise NotImplementedError(
+                "state_dict hasn't been implemented for this backend."
+            )
         if verbose > 0:
             print(
                 "Epoch {}: saving model to {}-{} ...\n".format(
@@ -479,43 +530,28 @@ class Model(object):
 
     def restore(self, save_path, verbose=0):
         """Restore all variables from a disk file."""
-        # TODO: backend tensorflow
+        # TODO: backend tensorflow, pytorch
+        if backend_name != "tensorflow.compat.v1":
+            raise NotImplementedError(
+                "state_dict hasn't been implemented for this backend."
+            )
         if verbose > 0:
             print("Restoring model from {} ...\n".format(save_path))
         self.saver.restore(self.sess, save_path)
 
     def print_model(self):
         """Prints all trainable variables."""
-        # TODO: backend tensorflow
+        # TODO: backend tensorflow, pytorch
+        if backend_name != "tensorflow.compat.v1":
+            raise NotImplementedError(
+                "state_dict hasn't been implemented for this backend."
+            )
         variables_names = [v.name for v in tf.trainable_variables()]
         values = self.sess.run(variables_names)
         for k, v in zip(variables_names, values):
             print("Variable: {}, Shape: {}".format(k, v.shape))
             print(v)
 
-    def _run(
-        self, fetches, training, dropout, data_id, inputs, targets, auxiliary_vars
-    ):
-        """Runs one "step" of computation of tensors or callables in `fetches`."""
-        if backend_name == "tensorflow.compat.v1":
-            feed_dict = self.net.feed_dict(
-                training,
-                dropout,
-                data_id,
-                inputs,
-                targets,
-                auxiliary_vars,
-            )
-            return self.sess.run(fetches, feed_dict=feed_dict)
-        if backend_name == "tensorflow":
-            # TODO: Support training, dropout
-            outs = fetches(data_id, inputs, targets, auxiliary_vars)
-            return None if outs is None else [out.numpy() for out in outs]
-        if backend_name == "pytorch":
-            # TODO: Use torch.no_grad() in _test() and predict()
-            outs = fetches(data_id, inputs, targets)
-            return None if outs is None else [out.detach().numpy() for out in outs]
-
 
 class TrainState(object):
     def __init__(self):
